@@ -5,19 +5,24 @@
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE Rank2Types #-}
 
 -- | A postgresql backend for persistent.
 module Database.Persist.Postgresql
     ( withPostgresqlPool
     , withPostgresqlConn
     , createPostgresqlPool
+    , createPostgresqlPoolModified
     , module Database.Persist.Sql
     , ConnectionString
     , PostgresConf (..)
     , openSimpleConn
+    , tableName
+    , fieldName
     ) where
 
 import Database.Persist.Sql
+import Database.Persist.Sql.Util (dbIdColumnsEsc)
 import Data.Fixed (Pico)
 
 import qualified Database.PostgreSQL.Simple as PG
@@ -44,6 +49,7 @@ import Data.List (find, sort, groupBy)
 import Data.Function (on)
 import Data.Conduit
 import qualified Data.Conduit.List as CL
+import Control.Monad.Logger (MonadLogger, runNoLoggingT)
 
 import qualified Data.IntMap as I
 
@@ -53,13 +59,16 @@ import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
 import qualified Blaze.ByteString.Builder.Char8 as BBB
 
-import Data.Time.LocalTime (localTimeToUTC, utc)
 import Data.Text (Text)
 import Data.Aeson
-import Control.Monad (forM, mzero)
+import Data.Aeson.Types (modifyFailure)
+import Control.Monad (forM)
+import Data.Acquire (Acquire, mkAcquire, with)
 import System.Environment (getEnvironment)
 import Data.Int (Int64)
 import Data.Monoid ((<>))
+import Data.Time (utc, localTimeToUTC)
+
 -- | A @libpq@ connection string.  A simple example of connection
 -- string would be @\"host=localhost port=5432 user=test
 -- dbname=test password=test\"@.  Please read libpq's
@@ -74,7 +83,7 @@ type ConnectionString = ByteString
 -- finishes using it.  Note that you should not use the given
 -- 'ConnectionPool' outside the action since it may be already
 -- been released.
-withPostgresqlPool :: MonadIO m
+withPostgresqlPool :: (MonadBaseControl IO m, MonadLogger m, MonadIO m)
                    => ConnectionString
                    -- ^ Connection string to the database.
                    -> Int
@@ -84,37 +93,57 @@ withPostgresqlPool :: MonadIO m
                    -- ^ Action to be executed that uses the
                    -- connection pool.
                    -> m a
-withPostgresqlPool ci = withSqlPool $ open' ci
+withPostgresqlPool ci = withSqlPool $ open' (const $ return ()) ci
 
 
 -- | Create a PostgreSQL connection pool.  Note that it's your
--- responsability to properly close the connection pool when
+-- responsibility to properly close the connection pool when
 -- unneeded.  Use 'withPostgresqlPool' for an automatic resource
 -- control.
-createPostgresqlPool :: MonadIO m
+createPostgresqlPool :: (MonadIO m, MonadBaseControl IO m, MonadLogger m)
                      => ConnectionString
                      -- ^ Connection string to the database.
                      -> Int
                      -- ^ Number of connections to be kept open
                      -- in the pool.
                      -> m ConnectionPool
-createPostgresqlPool ci = createSqlPool $ open' ci
+createPostgresqlPool = createPostgresqlPoolModified (const $ return ())
 
+-- | Same as 'createPostgresqlPool', but additionally takes a callback function
+-- for some connection-specific tweaking to be performed after connection
+-- creation. This could be used, for example, to change the schema. For more
+-- information, see:
+--
+-- <https://groups.google.com/d/msg/yesodweb/qUXrEN_swEo/O0pFwqwQIdcJ>
+--
+-- Since 2.1.3
+createPostgresqlPoolModified
+    :: (MonadIO m, MonadBaseControl IO m, MonadLogger m)
+    => (PG.Connection -> IO ()) -- ^ action to perform after connection is created
+    -> ConnectionString -- ^ Connection string to the database.
+    -> Int -- ^ Number of connections to be kept open in the pool.
+    -> m ConnectionPool
+createPostgresqlPoolModified modConn ci = createSqlPool $ open' modConn ci
 
 -- | Same as 'withPostgresqlPool', but instead of opening a pool
 -- of connections, only one connection is opened.
-withPostgresqlConn :: (MonadIO m, MonadBaseControl IO m)
-                   => ConnectionString -> (Connection -> m a) -> m a
-withPostgresqlConn = withSqlConn . open'
+withPostgresqlConn :: (MonadIO m, MonadBaseControl IO m, MonadLogger m)
+                   => ConnectionString -> (SqlBackend -> m a) -> m a
+withPostgresqlConn = withSqlConn . open' (const $ return ())
+
+open' :: (PG.Connection -> IO ())
+      -> ConnectionString -> LogFunc -> IO SqlBackend
+open' modConn cstr logFunc = do
+    conn <- PG.connectPostgreSQL cstr
+    modConn conn
+    openSimpleConn logFunc conn
 
-open' :: ConnectionString -> IO Connection
-open' cstr = PG.connectPostgreSQL cstr >>= openSimpleConn
 
 -- | Generate a 'Connection' from a 'PG.Connection'
-openSimpleConn :: PG.Connection -> IO Connection
-openSimpleConn conn = do
-    smap <- newIORef Map.empty
-    return Connection
+openSimpleConn :: LogFunc -> PG.Connection -> IO SqlBackend
+openSimpleConn logFunc conn = do
+    smap <- newIORef $ Map.empty
+    return SqlBackend
         { connPrepare    = prepare' conn
         , connStmtMap    = smap
         , connInsertSql  = insertSql'
@@ -127,6 +156,7 @@ openSimpleConn conn = do
         , connNoLimit    = "LIMIT ALL"
         , connRDBMS      = "postgresql"
         , connLimitOffset = decorateSQLWithLimitOffset "LIMIT ALL"
+        , connLogFunc = logFunc
         }
 
 prepare' :: PG.Connection -> Text -> IO Statement
@@ -139,7 +169,7 @@ prepare' conn sql = do
         , stmtQuery = withStmt' conn query
         }
 
-insertSql' :: EntityDef SqlType -> [PersistValue] -> InsertSqlResult
+insertSql' :: EntityDef -> [PersistValue] -> InsertSqlResult
 insertSql' ent vals =
   let sql = T.concat
                 [ "INSERT INTO "
@@ -156,18 +186,18 @@ insertSql' ent vals =
                 ]
   in case entityPrimary ent of
        Just _pdef -> ISRManyKeys sql vals
-       Nothing -> ISRSingle (sql <> " RETURNING " <> escape (entityID ent))
+       Nothing -> ISRSingle (sql <> " RETURNING " <> escape (fieldDB (entityId ent)))
 
 execute' :: PG.Connection -> PG.Query -> [PersistValue] -> IO Int64
 execute' conn query vals = PG.execute conn query (map P vals)
 
-withStmt' :: MonadResource m
+withStmt' :: MonadIO m
           => PG.Connection
           -> PG.Query
           -> [PersistValue]
-          -> Source m [PersistValue]
+          -> Acquire (Source m [PersistValue])
 withStmt' conn query vals =
-    bracketP openS closeS pull
+    pull `fmap` mkAcquire openS closeS
   where
     openS = do
       -- Construct raw query
@@ -188,11 +218,7 @@ withStmt' conn query vals =
                 status <- LibPQ.resultStatus ret
                 case status of
                   LibPQ.TuplesOk -> return ()
-                  _ -> do
-                    msg <- LibPQ.resStatus status
-                    mmsg <- LibPQ.resultErrorMessage ret
-                    fail $ "Postgresql.withStmt': bad result status " ++
-                           show status ++ " (" ++ maybe (show msg) (show . (,) msg) mmsg ++ ")"
+                  _ -> PG.throwResultError "Postgresql.withStmt': bad result status " ret status
 
                 -- Get number and type of columns
                 cols <- LibPQ.nfields ret
@@ -201,9 +227,8 @@ withStmt' conn query vals =
                 rowRef   <- newIORef (LibPQ.Row 0)
                 rowCount <- LibPQ.ntuples ret
                 return (ret, rowRef, rowCount, oids)
-      getters <- forM ids $ \(col, oid) -> do
-          getter <- getGetter conn oid
-          return $ getter $ PG.Field rt col oid
+      let getters
+            = map (\(col, oid) -> getGetter conn oid $ PG.Field rt col oid) ids
       return (rt, rr, rc, getters)
 
     closeS (ret, _, _, _) = LibPQ.unsafeFreeResult ret
@@ -245,7 +270,6 @@ instance PGTF.ToField P where
     toField (P (PersistDay d))         = PGTF.toField d
     toField (P (PersistTimeOfDay t))   = PGTF.toField t
     toField (P (PersistUTCTime t))     = PGTF.toField t
-    toField (P (PersistZonedTime (ZT t))) = PGTF.toField t
     toField (P PersistNull)            = PGTF.toField PG.Null
     toField (P (PersistList l))        = PGTF.toField $ listToJSON l
     toField (P (PersistMap m))         = PGTF.toField $ mapToJSON m
@@ -290,16 +314,18 @@ builtinGetters = I.fromList
     , (k PS.varchar,     convertPV PersistText)
     , (k PS.date,        convertPV PersistDay)
     , (k PS.time,        convertPV PersistTimeOfDay)
-    , (k PS.timestamp,   convertPV (PersistUTCTime . localTimeToUTC utc))
-    , (k PS.timestamptz, convertPV (PersistZonedTime . ZT))
+    , (k PS.timestamp,   convertPV (PersistUTCTime. localTimeToUTC utc))
+    , (k PS.timestamptz, convertPV PersistUTCTime)
     , (k PS.bit,         convertPV PersistInt64)
     , (k PS.varbit,      convertPV PersistInt64)
     , (k PS.numeric,     convertPV PersistRational)
     , (k PS.void,        \_ _ -> return PersistNull)
-    , (k PS.uuid,        convertPV (PersistDbSpecific . unUnknown))
     , (k PS.json,        convertPV (PersistByteString . unUnknown))
+    , (k PS.jsonb,       convertPV (PersistByteString . unUnknown))
     , (k PS.unknown,     convertPV (PersistByteString . unUnknown))
 
+
+
     -- array types: same order as above
     , (1000,             listOf PersistBool)
     , (1001,             listOf (PersistByteString . unBinary))
@@ -319,34 +345,25 @@ builtinGetters = I.fromList
     , (1015,             listOf PersistText)
     , (1182,             listOf PersistDay)
     , (1183,             listOf PersistTimeOfDay)
-    , (1115,             listOf (PersistUTCTime . localTimeToUTC utc))
-    , (1185,             listOf (PersistZonedTime . ZT))
+    , (1115,             listOf PersistUTCTime)
+    , (1185,             listOf PersistUTCTime)
     , (1561,             listOf PersistInt64)
     , (1563,             listOf PersistInt64)
     , (1231,             listOf PersistRational)
     -- no array(void) type
     , (2951,             listOf (PersistDbSpecific . unUnknown))
     , (199,              listOf (PersistByteString . unUnknown))
+    , (3807,             listOf (PersistByteString . unUnknown))
     -- no array(unknown) either
     ]
     where
         k (PGFF.typoid -> i) = PG.oid2int i
         listOf f = convertPV (PersistList . map f . PG.fromPGArray)
 
-getGetter :: PG.Connection -> PG.Oid -> IO (Getter PersistValue)
-getGetter conn oid = case I.lookup (PG.oid2int oid) builtinGetters of
-    Just getter -> return getter
-    Nothing -> do
-        tyinfo <- PG.getTypeInfo conn oid
-        error $ "Postgresql.getGetter: type "
-             ++ explain tyinfo
-             ++ " ("
-             ++ show oid
-             ++ ") not supported."
-    where
-        explain tyinfo = case B8.unpack $ PG.typname tyinfo of
-                        t@('_':ty) -> t ++ " (array of " ++ ty ++ ")"
-                        x -> show x
+getGetter :: PG.Connection -> PG.Oid -> Getter PersistValue
+getGetter _conn oid
+  = fromMaybe defaultGetter $ I.lookup (PG.oid2int oid) builtinGetters
+  where defaultGetter = convertPV (PersistDbSpecific . unUnknown)
 
 unBinary :: PG.Binary a -> a
 unBinary (PG.Binary x) = x
@@ -356,7 +373,7 @@ doesTableExist :: (Text -> IO Statement)
                -> IO Bool
 doesTableExist getter (DBName name) = do
     stmt <- getter sql
-    runResourceT $ stmtQuery stmt vals $$ start
+    with (stmtQuery stmt vals) ($$ start)
   where
     sql = "SELECT COUNT(*) FROM information_schema.tables WHERE table_name=?"
     vals = [PersistText name]
@@ -367,60 +384,87 @@ doesTableExist getter (DBName name) = do
     start' res = error $ "doesTableExist returned unexpected result: " ++ show res
     finish x = await >>= maybe (return x) (error "Too many rows returned in doesTableExist")
 
-migrate' :: [EntityDef a]
+migrate' :: [EntityDef]
          -> (Text -> IO Statement)
-         -> EntityDef SqlType
+         -> EntityDef
          -> IO (Either [Text] [(Bool, Text)])
-migrate' allDefs getter val = fmap (fmap $ map showAlterDb) $ do
-    let name = entityDB val
-    old <- getColumns getter val
+migrate' allDefs getter entity = fmap (fmap $ map showAlterDb) $ do
+    old <- getColumns getter entity
     case partitionEithers old of
         ([], old'') -> do
-            let old' = partitionEithers old''
-            let (newcols', udefs, fdefs) = mkColumns allDefs val
-            let newcols = filter (not . safeToRemove val . cName) newcols'
-            let udspair = map udToPair udefs
-            -- Check for table existence if there are no columns, workaround
-            -- for https://github.com/yesodweb/persistent/issues/152
             exists <-
                 if null old
                     then doesTableExist getter name
                     else return True
-
-            if not exists
-                then do
-                    let idtxt = case entityPrimary val of
-                                  Just pdef -> T.concat [" PRIMARY KEY (", T.intercalate "," $ map (escape . snd) $ primaryFields pdef, ")"]
-                                  Nothing   -> T.concat [escape $ entityID val
-                                        , " SERIAL PRIMARY KEY UNIQUE"]
-                    let addTable = AddTable $ T.concat
-                            -- Lower case e: see Database.Persist.Sql.Migration
-                            [ "CREATe TABLE "
-                            , escape name
-                            , "("
-                            , idtxt
-                            , if null newcols then "" else ","
-                            , T.intercalate "," $ map showColumn newcols
-                            , ")"
-                            ]
-                    let uniques = flip concatMap udspair $ \(uname, ucols) ->
-                            [AlterTable name $ AddUniqueConstraint uname ucols]
-                        references = mapMaybe (\c@Column { cName=cname, cReference=Just (refTblName, _) } -> getAddReference allDefs name refTblName cname (cReference c)) $ filter (isJust . cReference) newcols
-                        foreignsAlt = map (\fdef -> let (childfields, parentfields) = unzip (map (\(_,b,_,d) -> (b,d)) (foreignFields fdef))
-                                                    in AlterColumn name (foreignRefTableDBName fdef, AddReference (foreignConstraintNameDBName fdef) childfields parentfields)) fdefs
-                    return $ Right $ addTable : uniques ++ references ++ foreignsAlt
-                else do
-                    let (acs, ats) = getAlters allDefs val (newcols, udspair) old'
-                    let acs' = map (AlterColumn name) acs
-                    let ats' = map (AlterTable name) ats
-                    return $ Right $ acs' ++ ats'
+            return $ Right $ migrationText exists old''
         (errs, _) -> return $ Left errs
+  where
+    name = entityDB entity
+    migrationText exists old'' =
+        if not exists
+            then createText newcols fdefs udspair
+            else let (acs, ats) = getAlters allDefs entity (newcols, udspair) old'
+                     acs' = map (AlterColumn name) acs
+                     ats' = map (AlterTable name) ats
+                 in  acs' ++ ats'
+       where
+         old' = partitionEithers old''
+         (newcols', udefs, fdefs) = mkColumns allDefs entity
+         newcols = filter (not . safeToRemove entity . cName) newcols'
+         udspair = map udToPair udefs
+            -- Check for table existence if there are no columns, workaround
+            -- for https://github.com/yesodweb/persistent/issues/152
+
+    createText newcols fdefs udspair =
+        addTable : uniques ++ references ++ foreignsAlt
+      where
+        addTable = AddTable $ T.concat
+                -- Lower case e: see Database.Persist.Sql.Migration
+                [ "CREATe TABLE " -- DO NOT FIX THE CAPITALIZATION!
+                , escape name
+                , "("
+                , idtxt
+                , if null newcols then "" else ","
+                , T.intercalate "," $ map showColumn newcols
+                , ")"
+                ]
+        uniques = flip concatMap udspair $ \(uname, ucols) ->
+                [AlterTable name $ AddUniqueConstraint uname ucols]
+        references = mapMaybe (\c@Column { cName=cname, cReference=Just (refTblName, _) } ->
+            getAddReference allDefs name refTblName cname (cReference c))
+                   $ filter (isJust . cReference) newcols
+        foreignsAlt = flip map fdefs (\fdef ->
+            let (childfields, parentfields) = unzip (map (\((_,b),(_,d)) -> (b,d)) (foreignFields fdef))
+            in AlterColumn name (foreignRefTableDBName fdef, AddReference (foreignConstraintNameDBName fdef) childfields (map escape parentfields)))
+
+    idtxt = case entityPrimary entity of
+                Just pdef -> T.concat [" PRIMARY KEY (", T.intercalate "," $ map (escape . fieldDB) $ compositeFields pdef, ")"]
+                Nothing   ->
+                    let defText = defaultAttribute $ fieldAttrs $ entityId entity
+                        sType = fieldSqlType $ entityId entity
+                    in  T.concat
+                            [ escape $ fieldDB (entityId entity)
+                            , maySerial sType defText
+                            , " PRIMARY KEY UNIQUE"
+                            , mayDefault defText
+                            ]
+
+
+maySerial :: SqlType -> Maybe Text -> Text
+maySerial SqlInt64 Nothing = " SERIAL "
+maySerial sType _ = " " <> showSqlType sType
+
+mayDefault :: Maybe Text -> Text
+mayDefault def = case def of
+    Nothing -> ""
+    Just d -> " DEFAULT " <> d
 
 type SafeToRemove = Bool
 
-data AlterColumn = Type SqlType | IsNull | NotNull | Add' Column | Drop SafeToRemove
+data AlterColumn = Type SqlType Text
+                 | IsNull | NotNull | Add' Column | Drop SafeToRemove
                  | Default Text | NoDefault | Update' Text
-                 | AddReference DBName [DBName] [DBName] | DropReference DBName
+                 | AddReference DBName [DBName] [Text] | DropReference DBName
 type AlterColumn' = (DBName, AlterColumn)
 
 data AlterTable = AddUniqueConstraint DBName [DBName]
@@ -432,7 +476,7 @@ data AlterDB = AddTable Text
 
 -- | Returns all of the columns in the given table currently in the database.
 getColumns :: (Text -> IO Statement)
-           -> EntityDef a
+           -> EntityDef
            -> IO [Either Text (Either Column (DBName, [DBName]))]
 getColumns getter def = do
     let sqlv=T.concat ["SELECT "
@@ -451,9 +495,9 @@ getColumns getter def = do
     stmt <- getter sqlv
     let vals =
             [ PersistText $ unDBName $ entityDB def
-            , PersistText $ unDBName $ entityID def
+            , PersistText $ unDBName $ fieldDB (entityId def)
             ]
-    cs <- runResourceT $ stmtQuery stmt vals $$ helper
+    cs <- with (stmtQuery stmt vals) ($$ helper)
     let sqlc = T.concat ["SELECT "
                           ,"c.constraint_name, "
                           ,"c.column_name "
@@ -472,7 +516,7 @@ getColumns getter def = do
 
     stmt' <- getter sqlc
 
-    us <- runResourceT $ stmtQuery stmt' vals $$ helperU
+    us <- with (stmtQuery stmt' vals) ($$ helperU)
     return $ cs ++ us
   where
     getAll front = do
@@ -500,14 +544,14 @@ getColumns getter def = do
 
 -- | Check if a column name is listed as the "safe to remove" in the entity
 -- list.
-safeToRemove :: EntityDef a -> DBName -> Bool
+safeToRemove :: EntityDef -> DBName -> Bool
 safeToRemove def (DBName colName)
     = any (elem "SafeToRemove" . fieldAttrs)
     $ filter ((== DBName colName) . fieldDB)
     $ entityFields def
 
-getAlters :: [EntityDef a]
-          -> EntityDef SqlType
+getAlters :: [EntityDef]
+          -> EntityDef
           -> ([Column], [(DBName, [DBName])])
           -> ([Column], [(DBName, [DBName])])
           -> ([AlterColumn'], [AlterTable])
@@ -553,14 +597,23 @@ getColumn getter tname [PersistText x, PersistText y, PersistText z, d, npre, ns
                         { cName = cname
                         , cNull = y == "YES"
                         , cSqlType = t
-                        , cDefault = fmap
-                            (\xx -> fromMaybe xx $ T.stripSuffix "::character varying" xx)
-                                     d''
+                        , cDefault = fmap stripSuffixes d''
                         , cDefaultConstraintName = Nothing
                         , cMaxLen = Nothing
                         , cReference = ref
                         }
   where
+    stripSuffixes t =
+        loop'
+            [ "::character varying"
+            , "::text"
+            ]
+      where
+        loop' [] = t
+        loop' (p:ps) =
+            case T.stripSuffix p t of
+                Nothing -> loop' ps
+                Just t' -> t'
     getRef cname = do
         let sql = T.concat
                 [ "SELECT COUNT(*) FROM "
@@ -573,12 +626,12 @@ getColumn getter tname [PersistText x, PersistText y, PersistText z, d, npre, ns
                 ]
         let ref = refName tname cname
         stmt <- getter sql
-        runResourceT $ stmtQuery stmt
+        with (stmtQuery stmt
                      [ PersistText $ unDBName tname
                      , PersistText $ unDBName ref
-                     ] $$ do
+                     ]) ($$ do
             Just [PersistInt64 i] <- CL.head
-            return $ if i == 0 then Nothing else Just (DBName "", ref)
+            return $ if i == 0 then Nothing else Just (DBName "", ref))
     d' = case d of
             PersistNull   -> Right Nothing
             PersistText t -> Right $ Just t
@@ -588,8 +641,7 @@ getColumn getter tname [PersistText x, PersistText y, PersistText z, d, npre, ns
     getType "varchar"     = Right SqlString
     getType "date"        = Right SqlDay
     getType "bool"        = Right SqlBool
-    getType "timestamp"   = Right SqlDayTime
-    getType "timestamptz" = Right SqlDayTimeZoned
+    getType "timestamptz" = Right SqlDayTime
     getType "float4"      = Right SqlReal
     getType "float8"      = Right SqlReal
     getType "bytea"       = Right SqlBlob
@@ -607,7 +659,7 @@ sqlTypeEq :: SqlType -> SqlType -> Bool
 sqlTypeEq x y =
     T.toCaseFold (showSqlType x) == T.toCaseFold (showSqlType y)
 
-findAlters :: [EntityDef a] -> DBName -> Column -> [Column] -> ([AlterColumn'], [Column])
+findAlters :: [EntityDef] -> DBName -> Column -> [Column] -> ([AlterColumn'], [Column])
 findAlters defs _tablename col@(Column name isNull sqltype def _defConstraintName _maxLen ref) cols =
     case filter (\c -> cName c == name) cols of
         [] -> ([(name, Add' col)], cols)
@@ -615,9 +667,10 @@ findAlters defs _tablename col@(Column name isNull sqltype def _defConstraintNam
             let refDrop Nothing = []
                 refDrop (Just (_, cname)) = [(name, DropReference cname)]
                 refAdd Nothing = []
-                refAdd (Just (tname, a)) = case find ((==tname) . entityDB) defs of
-                                                Just refdef -> [(tname, AddReference a [name] [entityID refdef])]
-                                                Nothing -> error $ "could not find the entityDef for reftable[" ++ show tname ++ "]"
+                refAdd (Just (tname, a)) =
+                    case find ((==tname) . entityDB) defs of
+                        Just refdef -> [(tname, AddReference a [name] (dbIdColumnsEsc escape refdef))]
+                        Nothing -> error $ "could not find the entityDef for reftable[" ++ show tname ++ "]"
                 modRef =
                     if fmap snd ref == fmap snd ref'
                         then []
@@ -630,7 +683,18 @@ findAlters defs _tablename col@(Column name isNull sqltype def _defConstraintNam
                                             Just s -> (:) (name, Update' s)
                                  in up [(name, NotNull)]
                             _ -> []
-                modType = if sqlTypeEq sqltype sqltype' then [] else [(name, Type sqltype)]
+                modType
+                    | sqlTypeEq sqltype sqltype' = []
+                    -- When converting from Persistent pre-2.0 databases, we
+                    -- need to make sure that TIMESTAMP WITHOUT TIME ZONE is
+                    -- treated as UTC.
+                    | sqltype == SqlDayTime && sqltype' == SqlOther "timestamp" =
+                        [(name, Type sqltype $ T.concat
+                            [ " USING "
+                            , escape name
+                            , " AT TIME ZONE 'UTC'"
+                            ])]
+                    | otherwise = [(name, Type sqltype "")]
                 modDef =
                     if def == def'
                         then []
@@ -641,16 +705,16 @@ findAlters defs _tablename col@(Column name isNull sqltype def _defConstraintNam
                  filter (\c -> cName c /= name) cols)
 
 -- | Get the references to be added to a table for the given column.
-getAddReference :: [EntityDef a] -> DBName -> DBName -> DBName -> Maybe (DBName, DBName) -> Maybe AlterDB
+getAddReference :: [EntityDef] -> DBName -> DBName -> DBName -> Maybe (DBName, DBName) -> Maybe AlterDB
 getAddReference allDefs table reftable cname ref =
     case ref of
         Nothing -> Nothing
-        Just (s, _) -> Just $ AlterColumn table (s, AddReference (refName table cname) [cname] [id_])
+        Just (s, _) -> Just $ AlterColumn table (s, AddReference (refName table cname) [cname] id_)
                           where
                             id_ = fromMaybe (error $ "Could not find ID of entity " ++ show reftable)
                                         $ do
                                           entDef <- find ((== reftable) . entityDB) allDefs
-                                          return (entityID entDef)
+                                          return $ dbIdColumnsEsc escape entDef
 
 
 showColumn :: Column -> Text
@@ -673,8 +737,7 @@ showSqlType SqlReal = "DOUBLE PRECISION"
 showSqlType (SqlNumeric s prec) = T.concat [ "NUMERIC(", T.pack (show s), ",", T.pack (show prec), ")" ]
 showSqlType SqlDay = "DATE"
 showSqlType SqlTime = "TIME"
-showSqlType SqlDayTime = "TIMESTAMP"
-showSqlType SqlDayTimeZoned = "TIMESTAMP WITH TIME ZONE"
+showSqlType SqlDayTime = "TIMESTAMP WITH TIME ZONE"
 showSqlType SqlBlob = "BYTEA"
 showSqlType SqlBool = "BOOLEAN"
 
@@ -710,7 +773,7 @@ showAlterTable table (DropConstraint cname) = T.concat
     ]
 
 showAlter :: DBName -> AlterColumn' -> Text
-showAlter table (n, Type t) =
+showAlter table (n, Type t extra) =
     T.concat
         [ "ALTER TABLE "
         , escape table
@@ -718,6 +781,7 @@ showAlter table (n, Type t) =
         , escape n
         , " TYPE "
         , showSqlType t
+        , extra
         ]
 showAlter table (n, IsNull) =
     T.concat
@@ -786,7 +850,7 @@ showAlter table (reftable, AddReference fkeyname t2 id2) = T.concat
     , ") REFERENCES "
     , escape reftable
     , "("
-    , T.intercalate "," $ map escape id2
+    , T.intercalate "," id2
     , ")"
     ]
 showAlter table (_, DropReference cname) = T.concat
@@ -796,6 +860,22 @@ showAlter table (_, DropReference cname) = T.concat
     , escape cname
     ]
 
+-- | get the SQL string for the table that a PeristEntity represents
+-- Useful for raw SQL queries
+tableName :: forall record.
+          ( PersistEntity record
+          , PersistEntityBackend record ~ SqlBackend
+          ) => record -> Text
+tableName = escape . tableDBName
+
+-- | get the SQL string for the field that an EntityField represents
+-- Useful for raw SQL queries
+fieldName :: forall record typ.
+          ( PersistEntity record
+          , PersistEntityBackend record ~ SqlBackend
+          ) => EntityField record typ -> Text
+fieldName = escape . fieldDBName
+
 escape :: DBName -> Text
 escape (DBName s) =
     T.pack $ '"' : go (T.unpack s) ++ "\""
@@ -812,14 +892,11 @@ data PostgresConf = PostgresConf
       -- ^ The connection string.
     , pgPoolSize :: Int
       -- ^ How many connections should be held on the connection pool.
-    }
+    } deriving Show
 
-instance PersistConfig PostgresConf where
-    type PersistConfigBackend PostgresConf = SqlPersistT
-    type PersistConfigPool PostgresConf = ConnectionPool
-    createPoolConfig (PostgresConf cs size) = createPostgresqlPool cs size
-    runPool _ = runSqlPool
-    loadConfig (Object o) = do
+instance FromJSON PostgresConf where
+    parseJSON v = modifyFailure ("Persistent: error loadomg PostgreSQL conf: " ++) $
+      flip (withObject "PostgresConf") v $ \o -> do
         database <- o .: "database"
         host     <- o .: "host"
         port     <- o .:? "port" .!= 5432
@@ -835,7 +912,12 @@ instance PersistConfig PostgresConf where
                    }
             cstr = PG.postgreSQLConnectionString ci
         return $ PostgresConf cstr pool
-    loadConfig _ = mzero
+instance PersistConfig PostgresConf where
+    type PersistConfigBackend PostgresConf = SqlPersistT
+    type PersistConfigPool PostgresConf = ConnectionPool
+    createPoolConfig (PostgresConf cs size) = runNoLoggingT $ createPostgresqlPool cs size -- FIXME
+    runPool _ = runSqlPool
+    loadConfig = parseJSON
 
     applyEnv c0 = do
         env <- getEnvironment
